from hidi.transform import Transform


class DedupeTransform(Transform):
    """
    Deduplicate link-item tall skinny DataFrame
    """
    def __init__(self, skip_dedupe=False):
        self.skip_dedupe = skip_dedupe

    def transform(self, df, **kwargs):
        """
        Takes a :code:`df` that has :code:`link_id` and :code:`item_id`
        columns, and deduplicates them so that each pair is represented
        at most once.
        """
        if self.skip_dedupe:
            return df, kwargs

        return df.drop_duplicates(subset=['link_id', 'item_id']), kwargs



import numpy as np
import pandas as pd
import types
import warnings

from hidi.transform import Transform
from hidi.linalg import dot
from pyvalid import accepts
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
from sklearn.model_selection import train_test_split

# Catch annoying warnings from nimfa
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import nimfa as nf


class ApplyTransform(Transform):
    """
    Apply a function to an input.

    Takes a single argument, `fn`, which must be a function
    accepting one argument (the function to apply), and kwargs.
    """
    def __init__(self, fn):
        self.fn = fn

    def transform(self, x, **kwargs):
        return self.fn(x, **kwargs), kwargs


class SimilarityTransform(Transform):
    """
    Takes the dot product of a link*item matrix.

    Returns either a link*link or item*item similarity matrix.
    If axis is :code:`0`, an item*item matrix is returned, if
    axis is :code:`1` a link*link matrix is returned.
    The returned matrix represents a similarity matrix.

    The transform function returns a tuple containing the
    similarity matrix, and the links or items, depending on
    the axis.
    """

    def __init__(self, axis=0):
        self.axis = axis

    def transform(self, M, items, links, **kwargs):
        M_T = M.transpose()

        if self.axis == 0:
            sim_matrix = dot(M_T, M)
            sim_axis = items
        elif self.axis == 1:
            sim_matrix = dot(M, M_T)
            sim_axis = links
        else:
            raise Exception('Axis must be either 0 or 1')

        return sim_matrix, self.merge_kwargs(dict(items=sim_axis), kwargs)


class ScalarTransform(Transform):
    """
    Scale the matrix using a function or class method.

    `ScalerTransform` takes an `fn` argument that specifies the
    function that should be applied to the matrix. If `fn` is a string
    the scaler transform will try to call a function by that name on
    the matrix, if it is a function reference, scaler transform will
    call that function with the matrix as input.
    """

    def __init__(self, fn=np.log):
        self.fn = fn

    def scale(self, matrix):
        if isinstance(self.fn, types.FunctionType):
            return self.fn(matrix)
        elif type(self.fn) == str:
            return getattr(matrix, self.fn)()

        raise Exception('%s is not a valid scaling function' % self.fn)

    def transform(self, matrix_to_scale, **kwargs):
        """
        Takes a :code:`matrix_to_scale` as a numpy ndarray-like object
        and performs scaling on it, then returns the result.
        """
        out = self.scale(matrix_to_scale)

        return out, kwargs


class SparseTransform(Transform):
    """
    Make a sparse item*link matrix using SciPy's sparse compressed
    row matrix implementation.
    """

    @accepts(object, pd.DataFrame)
    def transform(self, df, **kwargs):
        """
        Takes a dataframe that has :code:`link_id`, :code:`item_id` and
        :code:`score` columns.

        Returns a SciPy :code:`csr_matrix`.
        """
        link_u = list(df.link_id.unique())
        item_u = list(df.item_id.unique())
        data = df.score.as_matrix()

        row = df.link_id.astype('category', categories=link_u).cat.codes
        col = df.item_id.astype('category', categories=item_u).cat.codes

        outshape = (len(link_u), len(item_u))
        in_tuple = (data, (row, col))
        kwargs = self.merge_kwargs(dict(links=link_u, items=item_u), kwargs)

        return csr_matrix(in_tuple, shape=outshape), kwargs


class DenseTransform(Transform):
    """
    Transform a sparse matrix to its dense representation.
    """
    def transform(self, M, **kwargs):
        return M.todense(), kwargs


class ItemsMatrixToDFTransform(Transform):
    """
    Create a Pandas DataFrame object with items as the index.
    """
    def transform(self, M, items, **kwargs):
        """
        Takes a numpy ndarray-like object and a list of item identifiers
        to be used as the index for the DataFrame.
        """
        return pd.DataFrame(M, index=items), kwargs


class KerasEvaluationTransform(Transform):
    """
    Generalized transform for Keras algorithm
    """
    def __init__(self, keras_model, validation_matrix, tts_seed=42,
                 tt_split=0.25, **keras_kwargs):
        self.keras_model = keras_model
        # seed, epochs, batch_size, verbose, cross_validation(boolean)
        self.keras_kwargs = keras_kwargs
        # labeled dataset for modeling and evaluation: item, labels
        self.validation_matrix = validation_matrix
        self.tts_seed = tts_seed
        self.tt_split = tt_split

        if 'item_id' in validation_matrix.columns:
            self.validation_matrix.set_index('item_id', inplace=True)

    def transform(self, M,  **kwargs):
        """
        Takes a numpy ndarray-like object and applies a Keras model to it.
        """
        # clean data
        rows, columns = M.shape
        embedding = M.merge(self.validation_matrix, left_index=True,
                            right_index=True)
        embedding = embedding.values

        # split dataset
        x_train, x_test, y_train, y_test = train_test_split(
            embedding[:, :columns], embedding[:, columns:],
            random_state=self.tts_seed, test_size=self.tt_split)

        self.keras_model.fit(
            x_train, y_train, validation_data=[x_test, y_test],
            **self.keras_kwargs)

        return self.keras_model, kwargs


class KerasPredictionTransform(Transform):
    """
    Generalized transform for Keras algorithm
    """
    def __init__(self, model):
        self.model = model
        # the model is the model output from KerasEvaluationTransform

    def transform(self, M,  **kwargs):
        """
        Takes a numpy ndarray-like object and applies a SkLearn
        algorithm to it.
        """
        predictions = self.model.predict(M)  # M is the ndarray-like object
        return predictions, kwargs


class SkLearnTransform(Transform):
    """
    Generalized transform for SciKit Learn algorithms.

    This transform takes a SciKit Learn algorithm, and its
    keyword arguments upon initialization. It applies the
    algorithm to the input when :code:`transform` is called.

    The algorithm to be applied is likely, but not necessarily
    a :code:`sklearn.decomposition` algorithm.
    """
    def __init__(self, SkLearnAlg, **sklearn_args):
        self.SkLearnAlg = SkLearnAlg
        self.sklearn_args = sklearn_args

    def transform(self, M, **kwargs):
        """
        Takes a numpy ndarray-like object and applies a SkLearn
        algorithm to it.
        """
        sklearn_alg = self.SkLearnAlg(**self.sklearn_args)
        transformed = sklearn_alg.fit_transform(M)
        kwargs['sklearn_fit'] = sklearn_alg

        return transformed, kwargs


class SVDTransform(SkLearnTransform):
    """
    Perform Truncated SVD on the matrix.

    This uses SciKit Learn's Tuncated SVD implementation, which
    is documented here:
    http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html

    All kwargs given to :code:`SVDTransform`'s initialization
    function will be given to :code:`sklearn.decomposition.TruncatedSVD`.

    Please reference the `sklearn docs
    <http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html>`_
    when using this transform.
    """
    def __init__(self, **svd_kwargs):
        super(SVDTransform, self).__init__(TruncatedSVD, **svd_kwargs)


class NimfaTransform(Transform):
    """
    Generalized Nimfa transform.

    This transform takes a nimfa algorithm, and its keyword
    arguments upon initialization. It applies the algorithm
    to the input when :code:`transform` is called.
    """
    def __init__(self, NimfaAlg, **nimfa_kwargs):
        self.NimfaAlg = NimfaAlg
        self.nimfa_kwargs = nimfa_kwargs

    def transform(self, M, **kwargs):
        nimfa_alg = self.NimfaAlg(M, **self.nimfa_kwargs)
        nimfa_fit = nimfa_alg()
        kwargs['nimfa_fit'] = nimfa_fit

        return nimfa_fit.basis(), kwargs


class SNMFTransform(NimfaTransform):
    """
    Perform Sparse Nonnegative Matrix Factorization.

    This wraps nimfa's snmf function, which is documented here:
    http://nimfa.biolab.si/nimfa.methods.factorization.snmf.html

    All kwargs given to :code:`SNFMTransform`'s initialization
    function will be given to :code:`nimfa.Snmf`.

    Please reference the `nimfa docs
    <http://nimfa.biolab.si/nimfa.methods.factorization.snmf.html>`_
    when using this transform.
    """

    def __init__(self, **snmf_kwargs):
        super(SNMFTransform, self).__init__(nf.Snmf, **snmf_kwargs)



import numpy as np
import pandas as pd

from hidi.transform import Transform


class ReadTransform(Transform):
    """
    Read input csv data from disk.

    Input data should be a csv file formatted with three
    columns: :code:`link_id`, :code:`item_id`, and
    :code:`score`. If score is not provided, it we be
    defaulted to one. :code:`link_id` represents to the
    "user" and `item_id` represents the "item" in the context
    of traditional collaborative filtering userXitem speak.
    """

    def __init__(self, infiles, **kwargs):
        self._inputs = infiles

    def _normalize(self, df):
        if 'score' not in df.columns:
            df['score'] = np.ones(df.shape[0])

        return df[['link_id', 'item_id', 'score']]

    def transform(self, **kwargs):
        """
        Read in files from the :code:`infiles` array given
        upon instantiation.
        """
        dfs = [pd.read_csv(inp) for inp in self._inputs]
        dfs = [self._normalize(df) for df in dfs]

        return pd.concat(dfs), kwargs


class WriteTransform(Transform):
    """
    Write output to disk in csv or json formats.

    Parameters:
        `outfile`: A string that is a path to the desired
        output on the file system.

        `file_format`: A string that is a file extension,
        either :code:`json` or :code:`csv`.
    """

    def __init__(self, outfile, file_format='csv',
                 enc=None, link_key='link_id'):
        self.outfile = outfile
        self.file_format = file_format
        self.link_key = link_key
        self.encoding = enc

    def transform(self, embeddings, **kwargs):
        if self.file_format == 'csv':
            embeddings.to_csv(self.outfile, encoding=self.encoding)
        else:
            with open(self.outfile, 'w+') as f:
                import json
                for row in embeddings.iterrows():
                    f.write(json.dumps({
                        self.link_key: row[0],
                        'embedding': row[1].tolist()
                    }))
                    f.write('\n')

        return embeddings, kwargs



class Transform(object):

    err = 'Tranform implementations must implement a transform function'

    def transform(self, *args):
        raise NotImplemented(self.err)

    def merge_kwargs(self, x, y):
        z = x.copy()
        z.update(y)

        return z



import numpy as np

from scipy import sparse


def dot(X, Y):
    if sparse.isspmatrix(X) and sparse.isspmatrix(Y):
        return X * Y
    elif sparse.isspmatrix(X) or sparse.isspmatrix(Y):
        return sparse.csr_matrix(X) * sparse.csr_matrix(Y)

    return np.asmatrix(X) * np.asmatrix(Y)



import sys

from codecs import open
from distutils.core import setup
from setuptools import find_packages


def get_install_requires():
    requires = [
        'scipy>=0.16.0',
        'tqdm>=4.0.0',
        'numpy>=1.8.0',
        'scikit-learn>=0.17.0',
        'pyvalid>=0.6',
        'pandas>=0.17.0',
        'nimfa>=1.1.0',
    ]

    if sys.version_info < (3, 2):
        requires.append('futures>=3.1.0')

    return requires


with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()


setup(
    name='HiDi',
    version='0.0.1',
    description='High-dimensional embedding generation library',
    long_description=readme,
    license='ALv2',
    author='Vevo Engineering',
    author_email='engineering@vevo.com',
    url='https://github.com/vevo/hidi',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=get_install_requires(),
    tests_require=[
        'nose==1.3.7',
        'Sphinx==1.5.5',
        'sphinx-autobuild==0.6.0',
        'tox==2.7.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4', 
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
)



from hidi import inout, clean, matrix, pipeline


# CSV file with link_id and item_id columns
in_files = ['hidi/examples/data/user-item.csv']

# File to write output data to
outfile = 'embeddings.csv'

transforms = [
    inout.ReadTransform(in_files),      # Read data from disk
    clean.DedupeTransform(),            # Dedupe it
    matrix.SparseTransform(),           # Make a sparse user*item matrix
    matrix.SimilarityTransform(),       # To item*item similarity matrix
    matrix.SVDTransform(),              # Perform SVD dimensionality reduction
    matrix.ItemsMatrixToDFTransform(),  # Make a DataFrame with an index
    inout.WriteTransform(outfile)       # Write results to csv
]

pl = pipeline.Pipeline(transforms)
pl.run()



import numpy as np

from hidi import pipeline, inout, matrix, forking


def to_float32(df, **kwargs):
    return df.astype(np.int32).astype(np.float32)


def create_pipeline(infiles):
    pl = pipeline.Pipeline([
        inout.ReadTransform(infiles),
        matrix.SparseTransform(),
        matrix.SimilarityTransform(),
        matrix.ApplyTransform(fn=to_float32),
        matrix.ScalarTransform(fn='log1p')
    ])

    left = pipeline.Pipeline([
        matrix.SNMFTransform(rank=32, max_iter=2),
        matrix.DenseTransform(),
        matrix.ItemsMatrixToDFTransform(),
        inout.WriteTransform('snmf-embeddings.csv')
    ])

    right = pipeline.Pipeline([
        matrix.SVDTransform(n_components=32, n_iter=2),
        matrix.ItemsMatrixToDFTransform(),
        inout.WriteTransform('svd-embeddings.csv')
    ])

    pl.add(forking.ProcessForkTransform([left, right], progress=False))

    return pl


def run_pipeline():
    pl = create_pipeline(['hidi/examples/data/user-item.csv'])

    return pl.run(progress=False)


if __name__ == '__main__':
    run_pipeline()



from tqdm import tqdm


class Pipeline(object):
    """
    Pipeline of transforms.

    Sequentially apply a list of transforms. All steps of the pipeline must be
    'transforms', that is, they must implement transform method. The Pipeline
    abstraction is inspired by the SciKit Learn Pipeline abstraction.

    Takes a list of transform instances.
    """

    def __init__(self, transformers):
        self.transformers = transformers
        self.fork_to = None
        self.fork_method = None

    def wrap_loop(self, to_iterate, progress):
        if progress:
            return tqdm(to_iterate, desc='Pipline Progress')

        return to_iterate

    def add(self, transform):
        """
        Add a transform to the pipeline.
        """
        self.transformers.append(transform)

    def run(self, io=None, progress=True, **kwargs):
        """
        Executes the pipeline and returns the final result.

        Takes an optional io parameter that will serve as input to the initial
        transformer.
        """
        for transformer in self.wrap_loop(self.transformers, progress):

            if io is not None:
                io, kwargs = transformer.transform(io, **kwargs)
            else:
                io, kwargs = transformer.transform(**kwargs)

        return io, kwargs



from hidi.transform import Transform
from functools import partial
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor


def map_fn(pipeline, io, progress, **kwargs):
    return pipeline.run(io, progress=progress, **kwargs)


class ExecutorFork(Transform):
    def __init__(self, pipelines, progress=False):
        self.pipelines = pipelines
        self.progress = progress

    def executor_fork(self, Executor, io, **kwargs):
        with Executor(len(self.pipelines)) as ex:
            run = partial(map_fn, io=io, progress=self.progress, **kwargs)

            return list(ex.map(run, self.pipelines)), kwargs


class ThreadForkTransform(ExecutorFork):
    """
    Fork a pipeline using :code:`concurrent.futures.ThreadPoolExecutor`
    as a backend for execution.

    This transform takes a list of Pipeline instances upon
    initialization.

    This is useful if you have several transforms that perform well
    when running in concurrent threads such as IO heavy or CPU heavy
    tasks that execute outside the Python runtime.

    The forked transform will return a list of Pipeline outputs,
    in the same order as the forked pipelines were given.
    """
    def transform(self, io, **kwargs):
        return self.executor_fork(ThreadPoolExecutor, io, **kwargs)


class ProcessForkTransform(ExecutorFork):
    """
    Fork a pipeline using :code:`concurrent.futures.ProcessesPoolExecutor`
    as a backend for execution.

    This transform takes a list of Pipeline instances upon initialization.

    This method is useful if you have several transforms that
    can be executed concurrently and are CPU intensive.

    The forked pipeline will now return a list of pipeline ouputs,
    in the same order as the forked pipelines were given.

    Special care must be taken as each transform must be pickled
    to a new process.
    """
    def transform(self, io, **kwargs):
        return self.executor_fork(ProcessPoolExecutor, io, **kwargs)


class TrivialForkTransform(Transform):
    """
    Trivial Fork Transform using an ordinary loop.
    """
    def __init__(self, pipelines, progress=False):
        self.pipelines = pipelines
        self.progress = progress

    def transform(self, io, **kwargs):
        output = []

        for pipeline in self.pipelines:
            output.append(pipeline.run(io, progress=self.progress, **kwargs))

        return output, kwargs
